/*
 * Copyright (c) 2020-2020,2023-2024 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                           Gianluca Corsini (LAAS-CNRS)
 *                                Lorenzo Casalino (External Contributor)
 *                                                     on Mon Nov 23 2020
 */
#include "aclibdynamixel.h"

#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wordexp.h>

#include "private.h"


/* motor models */
extern const struct ldx_mem_s ldx_mx28t_mem;
extern const struct ldx_mem_s ldx_mx64t_mem;
extern const struct ldx_mem_s ldx_magnet_mem;

static const struct ldx_mem_s * const dxmodels[] = {
  [DX_MX28T] = &ldx_mx28t_mem,
  [DX_MX64T] = &ldx_mx64t_mem,
  [DX_MAGNET] = &ldx_magnet_mem
};


static inline uint32_t	ldx_mem_data(const struct ldx_mem_s *mem,
                                uint16_t addr);
static inline uint16_t	ldx_mem_idx(const struct ldx_mem_s *mem,
                                uint16_t addr);


/* --- ldx_init_mem -------------------------------------------------------- */

/* Return NULL and sets errno in case of errors */

struct ldx_mem_s *
ldx_init_mem(const char *path, enum dx_model model, uint8_t id)
{
  struct ldx_mem_s *mem;
  unsigned long value;
  char buf[1024];
  wordexp_t w;
  char *t, *p;
  FILE *fp;
  int i, n;

  /* dynamixel motor model selection */
  if (model <= 0 || model >= sizeof(dxmodels)/sizeof(*dxmodels)) {
    warnx("out-of-range motor model 0 <= %d < %zu",
          model, sizeof(dxmodels)/sizeof(*dxmodels));
    return errno = ERANGE, NULL;
  }
  if (!dxmodels[model]) {
    warnx("invalid motor model `%d'", model);
    return errno = EINVAL, NULL;
  }

  /* intialize memory */
  mem = malloc(sizeof(*mem));
  if (!mem) return NULL;

  memcpy(mem, dxmodels[model], sizeof(*mem));
  mem->data = malloc(mem->len * sizeof(*mem->data));
  if (!mem->data) { free(mem); return NULL; }

  for(i = 0; i < mem->len; i++) mem->data[i] = mem->map[i].dflt;

  /* build index table */
  for(i = 0; i < sizeof(mem->idx)/sizeof(*mem->idx); i++)
    mem->idx[i] = LDX_IDX_INVALID;
  for(i = 0; i < mem->len; i++) {
    if (mem->map[i].addr >= sizeof(mem->idx)/sizeof(*mem->idx)) {
      warnx("invalid address `%d'", mem->map[i].addr);
      return errno = E2BIG, NULL;
    }

    mem->idx[mem->map[i].addr] = i;
  }

  /* need at least DX_ID defined */
  if (ldx_mem_idx(mem, DX_ID) == LDX_IDX_INVALID) {
      warnx("missing address `%d'", DX_ID);
      return errno = ENOENT, NULL;
  }

  /* load eeprom from file if specified */
  if (path && *path) {
    if (wordexp(path, &w, WRDE_NOCMD)) {
      warn("%s", path); return NULL;
    }
    if (w.we_wordc != 1) {
      warnx("%s: invalid path", path); return errno = EINVAL, NULL;
    }
    mem->path = strdup(w.we_wordv[0]);
    wordfree(&w);

    fp = fopen(mem->path, "r");
    if (fp == NULL && errno != ENOENT) { /* ENOENT is not a failure */
      warn("%s", path); return NULL;
    }

    n = 0;
    while(fp && fgets(buf, sizeof(buf), fp)) {
      /* first token: label */
      t = strtok_r(buf, " \t\n\r", &p);
      if (!t) /* skip empty lines */ continue;
      if (*t == '#') /* skip comments */ continue;

      /* search label, starting from last match to hopefully save iterations */
      for(i = 0; i < mem->len; i++) {
        if (!strcmp(mem->map[n].label, t)) break;
        n = (n + 1) % mem->len;
      }
      if (i >= mem->len) {
        warnx("%s: invalid entry `%s'", path, t);
        continue;
      } else if (i >= mem->rambase) {
        warnx("%s: RAM entry `%s'", path, t);
        continue;
      }

      /* second token: value */
      t = strtok_r(NULL, " \t\n\r", &p);
      if (!t) /* skip lines with no value */ continue;
      if (*t == '#') /* skip comments */ continue;

      errno = 0;
      value = strtoul(t, &p, 0);
      if (*p || errno) {
        warnx("%s: invalid value `%s' for `%s'", path, t, mem->map[n].label);
        continue;
      }
      mem->data[n] = value;
      n = (n + 1) % mem->len;
    }

    if (fp) fclose(fp);
  }

  if (id != DX_DEFAULT_ID) {
    /* overwrite mem id with input argument id */
    mem->data[mem->idx[DX_ID]] = id;
  }

  /* set default path for saving eeprom, if not specified */
  if (!mem->path) {
    snprintf(
      buf, sizeof(buf), "/tmp/dx_eeprom_id%d", ldx_mem_data(mem, DX_ID));
    mem->path = strdup(buf);
  }
  if (ldx_save_eeprom(mem)) {
    free(mem);
    return NULL;
  };

  return mem;
}


/* --- ldx_reset_mem ------------------------------------------------------- */

void
ldx_reset_mem(struct ldx_mem_s *mem)
{
  int i;

  /* find rambase index */
  for(i = mem->rambase;
      i < sizeof(mem->idx)/sizeof(*mem->idx) && mem->idx[i] == LDX_IDX_INVALID;
      i++)
    /* empty body */;
  if (i >= sizeof(mem->idx)/sizeof(*mem->idx)) return;

  /* reset RAM, skipping EEPROM */
  for(i = mem->idx[i]; i < mem->len; i++)
    mem->data[i] = mem->map[i].dflt;
}


/* --- ldx_fini_mem -------------------------------------------------------- */

void
ldx_fini_mem(struct ldx_mem_s *mem)
{
  if (!mem) return;
  if (mem->data) free(mem->data);
  if (mem->path) free(mem->path);
  free(mem);
}


/* --- ldx_save_eeprom ----------------------------------------------------- */

int
ldx_save_eeprom(struct ldx_mem_s *mem)
{
  int i;
  FILE *fp;

  fp = fopen(mem->path, "w");
  if (fp == NULL) {
    warn("unable to save eeprom to `%s'", mem->path);
    return errno;
  }

  fprintf(fp, "# %s\n", ctime(&(time_t){time(NULL)}));
  for(i = 0; i < mem->len && mem->map[i].addr < mem->rambase; i++)
    fprintf(fp, "%s %u\n", mem->map[i].label, mem->data[i]);

  fclose(fp);
  return 0;
}

/* --- ldx_read_mem -------------------------------------------------------- */

int
ldx_read_mem(struct ldx_mem_s *mem, uint16_t addr, uint8_t len, uint8_t *bytes)
{
  uint16_t idx;
  int i;

  /* ensure starting address is valid */
  idx = ldx_mem_idx(mem, addr);
  if (idx == LDX_IDX_INVALID) {
    warnx("invalid address `0x%x'", addr);
    return ACCESS_ERR;
  }

  /* read [addr; addr + len] as little-endian bytes. Any holes due to
   * non-mapped addresses are filled with zeroes. */
  for(; len; idx = ldx_mem_idx(mem, addr)) {
    /* non-mapped address */
    if (idx == LDX_IDX_INVALID) {
      *bytes++ = 0;
      len--; addr++;
      continue;
    }

    /* write-only address */
    if (mem->map[idx].perm == LDX_WO) {
      warnx("write-only address `0x%x' (%s)", addr, mem->map[idx].label);
      return ACCESS_ERR;
    }

    /* read sz little-endian bytes from current address */
    for(i = 0; i < mem->map[idx].sz && i < len; i++)
      *bytes++ = mem->data[idx] >> (8 * i);
    len -= i; addr += i;
  }

  return 0;
}


/* --- ldx_write_mem ------------------------------------------------------- */

int
ldx_write_mem(struct ldx_mem_s *mem, uint16_t addr, uint8_t len, uint8_t *bytes)
{
  uint16_t idx;
  int i;

  /* ensure starting address is valid */
  idx = ldx_mem_idx(mem, addr);
  if (idx == LDX_IDX_INVALID) {
    warnx("invalid address `0x%x'", addr);
    return ACCESS_ERR;
  }

  /* check for locked ROM */
  if (addr < mem->rambase && ldx_mem_data(mem, DX_TORQUE_ENABLE)) {
    warnx("locked ROM, address `0x%x' (%s)", addr, mem->map[idx].label);
    return ACCESS_ERR;
  }

  /* write [addr; addr + len] as little-endian bytes. Any holes due to
   * non-mapped addresses are ignored. */
  for(; len; idx = ldx_mem_idx(mem, addr)) {
    /* non-mapped address */
    if (idx == LDX_IDX_INVALID) {
      bytes++; len--; addr++;
      continue;
    }

    /* read-only address */
    if (mem->map[idx].perm == LDX_RO) {
      warnx("read-only address `0x%x' (%s)", addr, mem->map[idx].label);
      return ACCESS_ERR;
    }

    /* ensure full write */
    if (len < mem->map[idx].sz) {
      warnx("short write address `0x%x' (%s)", addr, mem->map[idx].label);
      return DATA_LENGTH_ERR;
    }

    /* write sz little-endian bytes from current address */
    mem->data[idx] = 0;
    for(i = 0; i < mem->map[idx].sz; i++)
      mem->data[idx] |= ((*bytes++) << (8 * i));
    len -= i; addr += i;
  }

  /* save EEPROM if modified */
  if (addr < mem->rambase)
    ldx_save_eeprom(mem);

  return 0;
}


/* --- ldx_mem_data/idx ---------------------------------------------------- */

static inline uint32_t
ldx_mem_data(const struct ldx_mem_s *mem, uint16_t addr)
{
  uint16_t idx = ldx_mem_idx(mem, addr);
  return idx == LDX_IDX_INVALID ? 0 : mem->data[idx];
}

static inline uint16_t
ldx_mem_idx(const struct ldx_mem_s *mem, uint16_t addr)
{
  return addr >= sizeof(mem->idx)/sizeof(*mem->idx) ?
    LDX_IDX_INVALID : mem->idx[addr];
}


/* --- dynamixel_get_eeprom_path ------------------------------------------- */

const char *
dynamixel_get_eeprom_path(const struct dynamixel_s *dynamixel)
{
  return dynamixel->mem->path;
}


/* --- dynamixel_get_id ---------------------------------------------------- */

uint8_t
dynamixel_get_id(const struct dynamixel_s *dynamixel)
{
  return ldx_mem_data(dynamixel->mem, DX_ID);
}


/* --- dynamixel_get_motor_ref --------------------------------------------- */

/* Reads from memory the motor setpoints, then updates internal structure
 * `motor->ref` of dynamixel and returns a pointer to it */
struct dynamixelstate_motor_s *
dynamixel_get_motor_ref(struct dynamixel_s *dynamixel)
{
  struct ldx_mem_s *mem = dynamixel->mem;
  struct dynamixelstate_motor_s *ref = &dynamixel->motor->ref;

  /* update motor reference points */
  ref->effort = (int16_t)ldx_mem_data(mem, DX_GOAL_CURRENT) *
                mem->map[ldx_mem_idx(mem, DX_GOAL_CURRENT)].unit;
  ref->velocity = (int32_t)ldx_mem_data(mem, DX_GOAL_VELOCITY) *
                  mem->map[ldx_mem_idx(mem, DX_GOAL_VELOCITY)].unit;
  ref->position = (int32_t)ldx_mem_data(mem, DX_GOAL_POSITION) *
                  mem->map[ldx_mem_idx(mem, DX_GOAL_POSITION)].unit;

  return ref;
}

/* --- dynamixel_set_motor_state ------------------------------------------- */

/* Updates internal structure `motor->state` of dynamixel with the provided
 * state, and then it writes it to memory */
void
dynamixel_set_motor_state(struct dynamixel_s *dynamixel,
                          struct dynamixelstate_motor_s *state)
{
  struct dynamixelstate_motor_s *curr_state = &dynamixel->motor->state;
  struct ldx_mem_s *mem = dynamixel->mem;
  uint16_t idx;

  /* update current motor state and associated memory entries */
  *curr_state = *state;

  idx = ldx_mem_idx(mem, DX_PRESENT_CURRENT);
  mem->data[idx] = (int16_t)(curr_state->effort / mem->map[idx].unit);

  idx = ldx_mem_idx(mem, DX_PRESENT_VELOCITY);
  mem->data[idx] = (int32_t)(curr_state->velocity / mem->map[idx].unit);

  idx = ldx_mem_idx(mem, DX_PRESENT_POSITION);
  mem->data[idx] = (int32_t)(curr_state->position / mem->map[idx].unit);
}


/* --- dynamixel_get_velocity_pi ------------------------------------------- */

/* Reads from memory the velocity-controller gains, then updates internal
 * structure `ctrl->vel_pi` of dynamixel and returns pointer to it */
struct pi_s *
dynamixel_get_velocity_pi(struct dynamixel_s *dynamixel)
{
  struct pi_s *vel_pi = &dynamixel->ctrl->vel_pi;
  struct ldx_mem_s *mem = dynamixel->mem;

  /* retrieve velocity pi gains */
  vel_pi->kP = (int16_t)ldx_mem_data(mem, DX_VELOCITY_P_GAIN) *
               mem->map[ldx_mem_idx(mem, DX_VELOCITY_P_GAIN)].unit;
  vel_pi->kI = (int16_t)ldx_mem_data(mem, DX_VELOCITY_I_GAIN) *
               mem->map[ldx_mem_idx(mem, DX_VELOCITY_I_GAIN)].unit;

  return vel_pi;
}

/* --- dynamixel_get_position_pidff ---------------------------------------- */

/* Reads from memory the position-controller gains, then updates internal
 * structure `ctrl->pos_pidff` of dynamixel and returns pointer to it */
struct pidff_s *
dynamixel_get_position_pidff(struct dynamixel_s *dynamixel)
{
  struct pidff_s *pos_pidff = &dynamixel->ctrl->pos_pidff;
  struct ldx_mem_s *mem = dynamixel->mem;

  /* retrieve position pidff gains */
  pos_pidff->kP = (int16_t)ldx_mem_data(mem, DX_POSITION_P_GAIN) *
                  mem->map[ldx_mem_idx(mem, DX_POSITION_P_GAIN)].unit;
  pos_pidff->kI = (int16_t)ldx_mem_data(mem, DX_POSITION_I_GAIN) *
                  mem->map[ldx_mem_idx(mem, DX_POSITION_I_GAIN)].unit;
  pos_pidff->kD = (int16_t)ldx_mem_data(mem, DX_POSITION_D_GAIN) *
                  mem->map[ldx_mem_idx(mem, DX_POSITION_D_GAIN)].unit;
  pos_pidff->kFF1 = (int16_t)ldx_mem_data(mem, DX_POSITION_KFF1_GAIN) *
                    mem->map[ldx_mem_idx(mem, DX_POSITION_KFF1_GAIN)].unit;
  pos_pidff->kFF2 = (int16_t)ldx_mem_data(mem, DX_POSITION_KFF2_GAIN) *
                    mem->map[ldx_mem_idx(mem, DX_POSITION_KFF2_GAIN)].unit;

  return pos_pidff;
}

/* --- dynamixel_get_op_mode ----------------------------------------------- */

/* Reads from memory the operating mode, then returns that value */
enum dx_mode
dynamixel_get_op_mode(struct dynamixel_s *dynamixel)
{
  return (enum dx_mode)ldx_mem_data(dynamixel->mem, DX_OPERATING_MODE);
}
