/*
 * Copyright (c) 2020-2020,2023-2024 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                           Gianluca Corsini (LAAS-CNRS)
 *                                Lorenzo Casalino (External Contributor)
 *                                                     on Mon Nov 25 2020
 */
#include "aclibdynamixel.h"

#include <math.h>

#include "../private.h"

/* Source: Dynamixel MX-28T/R/AT/AR(2.0):
 * https://emanual.robotis.com/docs/en/dxl/mx/mx-28-2/
 */

static const struct ldx_mem_entry_s map[] = {
  { .addr =   0, .label = "model_number",
    .dflt = 30, .unit = 0, .sz = 2, .perm = LDX_RO },

  { .addr =   2, .label = "model_info",
    .dflt = 0, .unit = 0, .sz = 4, .perm = LDX_RO },

  { .addr =   6, .label = "firmware_ver",
    .dflt = 0, .unit = 0, .sz = 1, .perm = LDX_RO },

  { .addr =   7, .label = "id",
    .dflt = 1, .unit = 0, .sz = 1, .perm = LDX_RW },

  { .addr =   9, .label = "return_delay_time",
    .dflt = 250, .unit = 2./1000000, .sz = 1, .perm = LDX_RW },

  { .addr =  11, .label = "operating_mode",
    .dflt = 3, .unit = 0, .sz = 1, .perm = LDX_RW },

  { .addr =  12, .label = "secondary_id",
    .dflt = 255, .unit = 0, .sz = 1, .perm = LDX_RW },

  { .addr =  20, .label = "homing_offset",
    .dflt = 0, .unit = 0.088*M_PI/180, .sz = 4, .perm = LDX_RW },

  { .addr =  31, .label = "temperature_limit",
    .dflt = 80, .unit = 1, .sz = 1, .perm = LDX_RW },

  { .addr =  32, .label = "max_voltage_limit",
    .dflt = 160, .unit = 0.1, .sz = 2, .perm = LDX_RW },

  { .addr =  34, .label = "min_voltage_limit",
    .dflt = 95, .unit = 0.1, .sz = 2, .perm = LDX_RW },

  { .addr =  36, .label = "pwm_limit",
    .dflt = 885, .unit = 885*0.113/100, .sz = 2, .perm = LDX_RW },

  { .addr =  40, .label = "acceleration_limit",
    .dflt = 32767, .unit = 214.577*2*M_PI/3600, .sz = 4, .perm = LDX_RW },

  { .addr =  44, .label = "velocity_limit",
    .dflt = 230, .unit = 0.229*2*M_PI/60, .sz = 4, .perm = LDX_RW },

  { .addr =  64, .label = "torque_enable",
    .dflt = 0, .unit = 0, .sz = 1, .perm = LDX_RW },

  { .addr =  65, .label = "led",
    .dflt = 0, .unit = 0, .sz = 1, .perm = LDX_RW },

  { .addr =  68, .label = "status_return_level",
    .dflt = 2, .unit = 0, .sz = 1, .perm = LDX_RW },

  { .addr =  76, .label = "velocity_I_gain",
    .dflt = 1920, .unit = 1./65536, .sz = 2, .perm = LDX_RW },

  { .addr =  78, .label = "velocity_P_gain",
    .dflt = 100, .unit = 1./128, .sz = 2, .perm = LDX_RW },

  { .addr =  80, .label = "position_D_gain",
    .dflt = 0, .unit = 1./16, .sz = 2, .perm = LDX_RW },

  { .addr =  82, .label = "position_I_gain",
    .dflt = 0, .unit = 1./65536, .sz = 2, .perm = LDX_RW },

  { .addr =  84, .label = "position_P_gain",
    .dflt = 850, .unit = 1./128, .sz = 2, .perm = LDX_RW },

  { .addr =  88, .label = "position_KFF1_gain",
    .dflt = 0, .unit = 1./4, .sz = 2, .perm = LDX_RW },

  { .addr =  90, .label = "position_KFF2_gain",
    .dflt = 0, .unit = 1./4, .sz = 2, .perm = LDX_RW },

  { .addr = 104, .label = "goal_velocity",
    .dflt = 0, .unit = 0.229*2*M_PI/60, .sz = 4, .perm = LDX_RW },

  { .addr = 108, .label = "profile_acceleration",
    .dflt = 0, .unit = 0, .sz = 4, .perm = LDX_RW },

  { .addr = 112, .label = "profile_velocity",
    .dflt = 0, .unit = 0, .sz = 4, .perm = LDX_RW },

  { .addr = 116, .label = "goal_position",
    .dflt = 0, .unit = 0.088*M_PI/180, .sz = 4, .perm = LDX_RW },

  { .addr = 126, .label = "present_current",
    .dflt = 0, .unit = 1000*0.1/100, .sz = 2, .perm = LDX_RO },

  { .addr = 128, .label = "present_velocity",
    .dflt = 0, .unit = 0.229*2*M_PI/60, .sz = 4, .perm = LDX_RO },

  { .addr = 132, .label = "present_position",
    .dflt = 0, .unit = 0.088*M_PI/180, .sz = 4, .perm = LDX_RO }
};

const struct ldx_mem_s ldx_mx28t_mem = {
  .map = map,
  .len = sizeof(map)/sizeof(*map),
  .rambase = 0x40
};
