/*
 * Copyright (c) 2020-2020,2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Gianluca Corsini on Mon Nov 23 2020
 */
#ifndef H_LDX_PRIVATE
#define H_LDX_PRIVATE

/* Errors of Dynamixel Protocol 2.0
 * [https://emanual.robotis.com/docs/en/dxl/protocol2/#error]
 */
#define OFFSET_FREE_AREA_EERNO -255
#define RESULT_FAIL_ERR (OFFSET_FREE_AREA_EERNO + 1)
#define INST_ERR (OFFSET_FREE_AREA_EERNO + 2)
#define CRC_ERR (OFFSET_FREE_AREA_EERNO + 3)
#define DATA_RANGE_ERR (OFFSET_FREE_AREA_EERNO + 4)
#define DATA_LENGTH_ERR (OFFSET_FREE_AREA_EERNO + 5)
#define DATA_LIMIT_ERR (OFFSET_FREE_AREA_EERNO + 6)
#define ACCESS_ERR (OFFSET_FREE_AREA_EERNO + 7)

#include <stdbool.h>
#include <stdint.h>
#include <time.h>

#include "libdynamixel.h"

enum dx_ref {
  DX_MODEL_NUMBER =		0x00,
  DX_FIRMWARE_VERSION =		0x06,
  DX_ID =			0x07,
  DX_RETURN_DELAY_TIME =	0x09,
  DX_OPERATING_MODE =		0x0b,
  DX_SECONDARY_ID =		0x0c,
  DX_HOMING_OFFSET =		0x14,
  DX_TEMP_LIMIT =		0x1f,
  DX_MAX_VOLTAGE_LIMIT =	0x20,
  DX_MIN_VOLTAGE_LIMIT =	0x22,
  DX_PWM_LIMIT =		0x24,
  DX_CURRENT_LIMIT =		0x26,
  DX_ACCELERATION_LIMIT =	0x28,
  DX_VELOCITY_LIMIT =		0x2c,
  DX_TORQUE_ENABLE =		0x40,
  DX_LED =			0x41,
  DX_STATUS_RETURN_LEVEL =	0x44,
  DX_VELOCITY_I_GAIN =		0x4c,
  DX_VELOCITY_P_GAIN =		0x4e,
  DX_POSITION_D_GAIN =		0x50,
  DX_POSITION_I_GAIN =		0x52,
  DX_POSITION_P_GAIN =		0x54,
  DX_POSITION_KFF2_GAIN =	0x58,
  DX_POSITION_KFF1_GAIN =	0x5a,
  DX_GOAL_CURRENT =		0x66,
  DX_GOAL_VELOCITY =		0x68,
  DX_PROFILE_ACCELERATION =	0x6c,
  DX_PROFILE_VELOCITY =		0x70,
  DX_GOAL_POSITION =		0x74,
  DX_PRESENT_CURRENT =		0x7e,
  DX_PRESENT_VELOCITY = 0x80,
  DX_PRESENT_POSITION = 0x84
};


/* --- motor memory -------------------------------------------------------- */

enum ldx_perm {
  LDX_RO,
  LDX_WO,
  LDX_RW
};

enum ldx_idx {
  LDX_IDX_INVALID = (uint16_t)-1
};

struct ldx_mem_entry_s {
  uint16_t addr;
  const char *label;
  uint32_t dflt;
  double unit;
  enum ldx_perm perm;
  uint8_t sz;
};

struct ldx_mem_s {
  const struct ldx_mem_entry_s * const map;	/* memory mappings */
  const uint8_t len;				/* map len */
  const uint8_t rambase;			/* start of RAM region */
  uint16_t idx[1024];				/* addr index table */
  uint32_t *data;				/* runtime memory */
  uint8_t alert;
  bool error;
  char *path;
};


/* --- pty for communication protocol emulation  --------------------------- */

struct dynamixel_pty_s {
  int fd;			/* pty */
  char path[512];		/* symlink to ptsname */

  uint8_t buf[1024];		/* input ring buffer */
  uint16_t r, w;		/* read/write index */

  uint8_t state;		/* packet-parsing state  */
  uint32_t header;		/* last 4 bytes */
  uint16_t len, crc;		/* packet LEN field & checksum */
  uint8_t id, cmd;		/* motor ID, packet intruction */
  uint8_t p, params[256];	/* PARAMs len and values */
};


/* --- motor data ---------------------------------------------------------- */

struct dynamixel_ts_s {
  double dt;
  struct timespec now;
};

struct dynamixel_motor_s {
  struct dynamixelstate_motor_s state, ref;
};

struct dynamixel_ctrl_s {
  struct pi_s vel_pi;
  struct pidff_s pos_pidff;
};


/* --- simulation context -------------------------------------------------- */

struct dynamixel_s {
  struct ldx_mem_s * const mem;
  struct dynamixel_ts_s * const ts;
  struct dynamixel_motor_s * const motor;
  struct dynamixel_ctrl_s * const ctrl;
  enum dx_model const model;
};


#ifdef __cplusplus
extern "C" {
#endif

  /* pty.c */
  ssize_t	ldx_recv_msg(struct dynamixel_pty_s *pty);
  int		ldx_send_status_msg(const struct dynamixel_pty_s *pty,
                        uint8_t id, bool alert, uint8_t error,
                        uint16_t nparams, uint8_t *params);

  /* mem.c */
  struct ldx_mem_s *
		ldx_init_mem(const char *path, enum dx_model model,
                        uint8_t id);
  void		ldx_reset_mem(struct ldx_mem_s *mem);
  void		ldx_fini_mem(struct ldx_mem_s *mem);
  int		ldx_save_eeprom(struct ldx_mem_s *mem);
  int		ldx_read_mem(struct ldx_mem_s *mem, uint16_t addr,
                        uint8_t len, uint8_t *bytes);
  int		ldx_write_mem(struct ldx_mem_s *mem, uint16_t addr,
                        uint8_t len, uint8_t *bytes);

#ifdef __cplusplus
}
#endif

#endif /* H_LDX_PRIVATE */
