/*
 * Copyright (c) 2020,2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                    Gianluca Corsini on Mon Nov 23 2020
 */
#include "aclibdynamixel.h"

#include <err.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "private.h"


/* --- dynamixel_init ------------------------------------------------------ */

struct dynamixel_s *
dynamixel_init(enum dx_model model, uint8_t id, const char *eeprom_path)
{
  struct dynamixel_s *dynamixel;
  struct ldx_mem_s *mem;

  mem = ldx_init_mem(eeprom_path, model, id);
  if (!mem) return NULL;

  dynamixel = malloc(sizeof(*dynamixel));
  if (!dynamixel) { ldx_fini_mem(mem); return NULL; }

  memcpy(dynamixel, &(struct dynamixel_s){
      .mem = mem,
      .ts =  (struct dynamixel_ts_s *) malloc(sizeof(*dynamixel->ts)),
      .motor = (struct dynamixel_motor_s *) malloc(sizeof(*dynamixel->motor)),
      .ctrl = (struct dynamixel_ctrl_s *) malloc(sizeof(*dynamixel->ctrl)),
      .model = model
    }, sizeof(*dynamixel));

  /* defaults */
  *dynamixel->ts = (struct dynamixel_ts_s){.dt = 0.  /* all zeroes */};
  *dynamixel->motor = (struct dynamixel_motor_s) {
    .state.effort = 0, .state.velocity = 0, .state.position = 0,
    .ref.effort = 0, .ref.velocity = 0, .ref.position = 0
  };
  *dynamixel->ctrl = (struct dynamixel_ctrl_s) {
    .vel_pi.kP = 0, .vel_pi.kI = 0,
    .pos_pidff.kP = 0, .pos_pidff.kI = 0, .pos_pidff.kD = 0,
    .pos_pidff.kFF1 = 0, .pos_pidff.kFF2 = 0
  };

  return dynamixel;
}


/* --- dynamixel_fini ------------------------------------------------------ */

void
dynamixel_fini(struct dynamixel_s *dynamixel)
{
  if (!dynamixel) return;

  /* free all allocated memory pointed by any pointer */
  ldx_fini_mem(dynamixel->mem);
  if (dynamixel->ts) free(dynamixel->ts);
  if (dynamixel->motor) free(dynamixel->motor);
  if (dynamixel->ctrl) free(dynamixel->ctrl);

  free(dynamixel);
}


/* --- dynamixel_get_msg --------------------------------------------------- */

int
dynamixel_get_msg(struct dynamixel_pty_s *pty)
{
  int s;

  /* read incoming data */
  s = ldx_recv_msg(pty);
  if (s < 0) {
    switch(errno) {
      case EAGAIN: /* no data */
      case EIO: /* not connected */
        return 0;

      default:
        warn("receiving from %s", pty->path);
        return -1;
    }
  }

  return s;
}


/* --- dynamixel_process_msg ----------------------------------------------- */

/* Process last received message.
 * Returns 0 on succes (even in case of no data, not connected), error code in
 * case of error.
 */
int
dynamixel_process_msg(const struct dynamixel_s *dynamixel,
                      struct dynamixel_pty_s *pty)
{
  struct ldx_mem_s *mem = dynamixel->mem;

  uint16_t addr, len;
  uint8_t id;
  int s;

  id = dynamixel_get_id(dynamixel);

  /* ignore msgs for others */
  if (pty->id != id && pty->id != DX_BROADCAST_ID)
    return 0;

  /* check CRC */
  if (pty->crc) return CRC_ERR;

  /* process message */
  switch(pty->cmd) {
    case DX_PING: {
      /* no parameters */
      uint8_t buf[3];
      s = ldx_read_mem(mem, DX_MODEL_NUMBER, 2, buf);
      if (!s) s = ldx_read_mem(mem, DX_FIRMWARE_VERSION, 1, buf + 2);
      if (!s) s = ldx_send_status_msg(pty, id, 0, 0, 3, buf);
      if (s) goto send_err;
      break;
    }

    case DX_READ:
      /* Parameter 1	Low-order byte from the starting address
       * Parameter 2	High-order byte from the starting address
       * Parameter 3	Low-order byte from the data length (X)
       * Parameter 4	High-order byte from the data length (X) */
      if (pty->p != 4) {
        warnx("bad DX_READ message on %s", pty->path);
        return RESULT_FAIL_ERR;
      }

      addr = (pty->params[1] << 8) | pty->params[0];
      len = (pty->params[3] << 8) | pty->params[2];
      {
        uint8_t buf[len];
        s = ldx_read_mem(mem, addr, len, buf);
        if (s) {
          warnx("error reading %d bytes at 0x%x", len, addr);
          return s;
        }
        s = ldx_send_status_msg(pty, id, 0, 0, len, buf);
        if (s) goto send_err;
      }
      break;

    case DX_SYNC_READ:
      /* Parameter 1	Low-order byte from the starting address
       * Parameter 2	High-order byte from the starting address
       * Parameter 3	Low-order byte from the data length(X)
       * Parameter 4	High-order byte from the data length(X)
       * Parameter 4+1	ID of the 1st Device
       * Parameter 4+2	ID of the 2nd Device
       *     …	           …
       * Parameter 4+Y	ID of the Y-th Device
       *
       * NB: in the online manual is stated that the number of ID devices is
       * equal to X, i.e. the number given by the two bytes of data length.
       * This is not true, as demonstrated by the Example 5.9.3.
       *  Data length X = number of bytes to read from memory.
       *  Y = number of Device IDs specified.
       *  4 + Y = LEN field (2 bytes) of the received packet, and Y != X
       */
      if (pty->p < 5) {
        warnx("bad DX_SYNC_READ message on %s", pty->path);
        return RESULT_FAIL_ERR;
      }

      /* ignore message if we are not in the list of IDs */
      if (!memchr(&pty->params[4], id, pty->p - 4))
        break;

      /* process */
      addr = (pty->params[1] << 8) | pty->params[0];
      len = (pty->params[3] << 8) | pty->params[2];
      {
        uint8_t buf[len];
        s = ldx_read_mem(mem, addr, len, buf);
        if (s) {
          warnx("error reading %d bytes at 0x%x", len, addr);
          return s;
        }
        s = ldx_send_status_msg(pty, id, 0, 0, len, buf);
        if (s) goto send_err;
      }
      break;

    case DX_WRITE: {
      /* Parameter 1	Low-order byte from the starting address
       * Parameter 2	High-order byte from the starting address
       * Parameter 2+1	First Byte
       * Parameter 2+2	Second Byte
       *     …             …
       * Parameter 2+X	X-th Byte
       */
      if (pty->p < 3) {
        warnx("bad DX_WRITE message on %s", pty->path);
        return RESULT_FAIL_ERR;
      }

      addr = (pty->params[1] << 8) | pty->params[0];
      s = ldx_write_mem(mem, addr, pty->p - 2, pty->params + 2);
      if (s < 0) {
        warnx("error writing %d bytes at 0x%x", pty->p - 2, addr);
        return s;
      }
      s = ldx_send_status_msg(pty, id, 0, 0, 0, NULL);
      if (s) goto send_err;
      break;
    }

    case DX_SYNC_WRITE: {
      /* Parameter 1   Low-order byte from the starting address
       * Parameter 2	High-order byte from the starting address
       * Parameter 3	Low-order byte from the data length(X)
       * Parameter 4	High-order byte from the data length(X)
       *
       * Parameter 5   [1st Device] ID
       * Parameter 5+1   [1st Device] 1st Byte
       * Parameter 5+2   [1st Device] 2nd Byte
       *     …	         [1st Device]…
       * Parameter 5+X   [1st Device] X-th Byte
       *
       * Parameter 6   [2nd Device] ID
       * Parameter 6+1   [2nd Device] 1st Byte
       * Parameter 6+2   [2nd Device] 2nd Byte
       *     …	         [2nd Device]…
       * Parameter 6+X   [2nd Device] X-th Byte
       */
      int i;

      if (pty->p < 6) {
        warnx("bad DX_SYNC_WRITE message on %s", pty->path);
        return RESULT_FAIL_ERR;
      }

      addr = (pty->params[1] << 8) | pty->params[0];
      len = (pty->params[3] << 8) | pty->params[2];
      if ((pty->p - 4) % (len + 1)) { /* see structure of packet */
        warnx("bad DX_SYNC_WRITE message on %s", pty->path);
        return RESULT_FAIL_ERR;
      }

      /* ignore message if we are not in the list of IDs */
      for(i = 4; i < pty->p; i += len+1)
        if (pty->params[i] == id) break;
      if (i >= pty->p) break;

      /* process */
      s = ldx_write_mem(mem, addr, len, &pty->params[i+1]);
      if (s) {
        warnx("error writing %d bytes at 0x%x", len, addr);
        return s;
      }
      s = ldx_send_status_msg(pty, id, 0, 0, 0, NULL);
      if (s) goto send_err;
      break;
    }

    case DX_REBOOT:
      /* no parameters. */

      /* save eeprom, reset RAM */
      s = ldx_save_eeprom(mem);
      ldx_reset_mem(mem);

      s = ldx_send_status_msg(pty, id, 0, 0, 1, &(uint8_t){0});
      if (s) goto send_err;
      break;

    default:
      warnx("undefined instruction %d", pty->cmd);
      return INST_ERR;
  }

  return 0;

send_err:
  warnx("error sending message to %s", pty->path);
  return -1;
}
